@@ -3,11 +3,10 @@ import copy
 import enum
 import json
 import logging
-import os
-from contextlib import suppress
 
-from django.conf import settings
-from django.contrib.postgres.fields import ArrayField, JSONField
+import jsonschema
+
+from django.contrib.postgres.fields import ArrayField
 from django.contrib.postgres.indexes import GinIndex
 from django.contrib.postgres.search import SearchVectorField
 from django.core.exceptions import PermissionDenied, ValidationError
@@ -15,13 +14,20 @@ from django.core.validators import RegexValidator
 from django.db import models, transaction
 
 from resolwe.flow.expression_engines.exceptions import EvaluationError
-from resolwe.flow.models.utils import DirtyError, fill_with_defaults, validate_schema
+from resolwe.flow.models.utils import (
+    DirtyError,
+    fill_with_defaults,
+    validate_schema,
+    validation_schema,
+)
 from resolwe.flow.utils import dict_dot, get_data_checksum, iterate_fields
+from resolwe.observers.protocol import post_permission_changed, pre_permission_changed
+from resolwe.permissions.models import PermissionObject, PermissionQuerySet
 from resolwe.permissions.utils import assign_contributor_permissions, copy_permissions
 
 from .base import BaseModel, BaseQuerySet
 from .descriptor import DescriptorSchema
-from .entity import Entity
+from .entity import Entity, EntityQuerySet
 from .secret import Secret
 from .storage import Storage
 from .utils import (
@@ -31,6 +37,7 @@ from .utils import (
     render_descriptor,
     render_template,
 )
+from .worker import Worker
 
 # Compatibility for Python < 3.5.
 if not hasattr(json, "JSONDecodeError"):
@@ -48,7 +55,7 @@ class HandleEntityOperation(enum.Enum):
     PASS = "PASS"
 
 
-class DataQuerySet(BaseQuerySet):
+class DataQuerySet(BaseQuerySet, PermissionQuerySet):
     """Query set for Data objects."""
 
     @staticmethod
@@ -186,15 +193,16 @@ class DataQuerySet(BaseQuerySet):
                 kind=DataDependency.KIND_SUBPROCESS,
             )
             # Data was from a workflow / spawned process
-            copy_permissions(subprocess_parent, obj)
+            if not obj.in_container():
+                copy_permissions(subprocess_parent, obj)
 
         # Entity, Collection assignment
         entity_operation = self._handle_entity(obj)
         self._handle_collection(obj, entity_operation=entity_operation)
 
-        # Permissions:
-        assign_contributor_permissions(obj)
-        copy_permissions(obj.collection, obj)
+        # Assign contributor permission only if Data is not in the container.
+        if not obj.in_container():
+            assign_contributor_permissions(obj)
 
         return obj
 
@@ -222,18 +230,39 @@ class DataQuerySet(BaseQuerySet):
         for data in self:
             data.move_to_collection(destination_collection)
 
+    def annotate_sample_path(self, path, annotation_name, value_to_label=False):
+        """Add annotation to the Entity QuerySet.
 
-class Data(BaseModel):
+        The annotation is the value of the field with the given name (in the
+        given group).
+
+        :attr group_name: the name of the group annotation field belongs to.
+        :attr field_name: the name of the annotation field.
+        :attr annotation_name: the name under which annotation will be stored.
+            When empty the name f"{group_name}_{field_name}" will be used.
+        :attr value_to_label: optionally annotate with label instead of the
+            value. Only applicable when the vocabulary on the field is given.
+        """
+        annotation_data = EntityQuerySet._prepare_annotation_data(
+            path,
+            annotation_name,
+            value_to_label,
+            outerref_entity_pk_path="entity_id",
+        )
+        return self.annotate(**annotation_data)
+
+
+class Data(BaseModel, PermissionObject):
     """Postgres model for storing data."""
 
     class Meta(BaseModel.Meta):
         """Data Meta options."""
 
         permissions = (
-            ("view_data", "Can view data"),
-            ("edit_data", "Can edit data"),
-            ("share_data", "Can share data"),
-            ("owner_data", "Is owner of the data"),
+            ("view", "Can view data"),
+            ("edit", "Can edit data"),
+            ("share", "Can share data"),
+            ("owner", "Is owner of the data"),
         )
 
         indexes = [
@@ -263,6 +292,7 @@ class Data(BaseModel):
     STATUS_ERROR = "ER"
     #: data object is in dirty state
     STATUS_DIRTY = "DR"
+    # Assumption (in listener): ordered from least to most problematic.
     STATUS_CHOICES = (
         (STATUS_UPLOADING, "Uploading"),
         (STATUS_RESOLVING, "Resolving"),
@@ -331,7 +361,7 @@ class Data(BaseModel):
     process_progress = models.PositiveSmallIntegerField(default=0)
 
     #: return code
-    process_rc = models.PositiveSmallIntegerField(blank=True, null=True)
+    process_rc = models.SmallIntegerField(blank=True, null=True)
 
     #: info log message
     process_info = ArrayField(models.CharField(max_length=255), default=list)
@@ -343,10 +373,10 @@ class Data(BaseModel):
     process_error = ArrayField(models.CharField(max_length=255), default=list)
 
     #: actual inputs used by the process
-    input = JSONField(default=dict)
+    input = models.JSONField(default=dict)
 
     #: actual outputs of the process
-    output = JSONField(default=dict)
+    output = models.JSONField(default=dict)
 
     #: total size of data's outputs in bytes
     size = models.BigIntegerField()
@@ -357,7 +387,7 @@ class Data(BaseModel):
     )
 
     #: actual descriptor
-    descriptor = JSONField(default=dict)
+    descriptor = models.JSONField(default=dict)
 
     #: indicate whether `descriptor` doesn't match `descriptor_schema` (is dirty)
     descriptor_dirty = models.BooleanField(default=False)
@@ -405,58 +435,15 @@ class Data(BaseModel):
     #: field used for full-text search
     search = SearchVectorField(null=True)
 
+    #: process requirements overrides
+    process_resources = models.JSONField(default=dict)
+
     def __init__(self, *args, **kwargs):
         """Initialize attributes."""
         super().__init__(*args, **kwargs)
         self._original_name = self.name
         self._original_output = self.output
 
-    def save_storage(self, instance, schema):
-        """Save basic:json values to a Storage collection."""
-        for field_schema, fields, path in iterate_fields(instance, schema, ""):
-            name = field_schema["name"]
-            value = fields[name]
-            if field_schema.get("type", "").startswith("basic:json:"):
-                if value and not self.pk:
-                    raise ValidationError(
-                        "Data object must be `created` before creating `basic:json:` fields"
-                    )
-
-                if isinstance(value, int):
-                    # already in Storage
-                    continue
-
-                if isinstance(value, str):
-                    file_path = self.location.get_path(filename=value)
-                    if os.path.isfile(file_path):
-                        try:
-                            with open(file_path) as file_handler:
-                                value = json.load(file_handler)
-                        except json.JSONDecodeError:
-                            with open(file_path) as file_handler:
-                                content = file_handler.read()
-                                content = content.rstrip()
-                                raise ValidationError(
-                                    "Value of '{}' must be a valid JSON, current: {}".format(
-                                        name, content
-                                    )
-                                )
-
-                existing_storage_pk = None
-                with suppress(KeyError):
-                    existing_storage_pk = dict_dot(self._original_output, path)
-
-                if isinstance(existing_storage_pk, int):
-                    self.storages.filter(pk=existing_storage_pk).update(json=value)
-                    fields[name] = existing_storage_pk
-                else:
-                    storage = self.storages.create(
-                        name="Storage for data id {}".format(self.pk),
-                        contributor=self.contributor,
-                        json=value,
-                    )
-                    fields[name] = storage.pk
-
     def resolve_secrets(self):
         """Retrieve handles for all basic:secret: fields on input.
 
@@ -531,6 +518,14 @@ class Data(BaseModel):
         if self.name != self._original_name:
             self.named_by_user = True
 
+        try:
+            jsonschema.validate(
+                self.process_resources, validation_schema("process_resources")
+            )
+        except jsonschema.exceptions.ValidationError as exception:
+            # Re-raise as Django ValidationError
+            raise ValidationError(exception.message)
+
         create = self.pk is None
         if create:
             fill_with_defaults(self.input, self.process.input_schema)
@@ -546,17 +541,14 @@ class Data(BaseModel):
 
             validate_schema(self.input, self.process.input_schema)
 
-        elif render_name:
-            self._render_name()
-
-        self.save_storage(self.output, self.process.output_schema)
-
-        if self.status != Data.STATUS_ERROR:
             hydrate_size(self)
             # If only specified fields are updated (e.g. in executor), size needs to be added
             if "update_fields" in kwargs:
                 kwargs["update_fields"].append("size")
 
+        elif render_name:
+            self._render_name()
+
         render_descriptor(self)
 
         if self.descriptor_schema:
@@ -581,9 +573,11 @@ class Data(BaseModel):
 
     def delete(self, *args, **kwargs):
         """Delete the data model."""
+        if hasattr(self, "worker"):
+            if self.worker.status not in Worker.FINAL_STATUSES:
+                self.worker.terminate()
         # Store ids in memory as relations are also deleted with the Data object.
         storage_ids = list(self.storages.values_list("pk", flat=True))
-
         super().delete(*args, **kwargs)
 
         Storage.objects.filter(pk__in=storage_ids, data=None).delete()
@@ -601,35 +595,38 @@ class Data(BaseModel):
             inherit_collection=inherit_collection,
         )[0]
 
-    def get_runtime_path(self, filename=None):
-        """Get the runtime directory of the executor.
-
-        That is the script that created this object. When filename is not
-        None return the path to the filename in the working directory of the
-        executor.
-        """
-        if self.is_duplicate():
-            raise Exception("Copied Data objects have no runtime directory.")
-        return self.location.get_path(
-            prefix=settings.FLOW_EXECUTOR["RUNTIME_DIR"], filename=filename
-        )
-
-    def move_to_collection(self, destination_collection):
+    def move_to_collection(self, collection):
         """Move data object to collection."""
-        self.validate_change_collection(destination_collection)
-        self.collection = destination_collection
-        if destination_collection:
-            self.tags = destination_collection.tags
-            copy_permissions(destination_collection, self)
-        self.save()
+        self.validate_change_collection(collection)
+        pre_permission_changed.send(sender=type(self), instance=self)
+        self.collection = collection
+        self.permission_group = collection.permission_group
+        if collection:
+            self.tags = collection.tags
+        self.save(update_fields=["tags", "permission_group", "collection"])
+        post_permission_changed.send(sender=type(self), instance=self)
 
-    def validate_change_collection(self, destination_collection):
+    def move_to_entity(self, entity):
+        """Move data object to entity."""
+        if entity is None and self.in_container():
+            raise ValidationError("Data object must not be removed from container.")
+        pre_permission_changed.send(sender=type(self), instance=self)
+        self.entity = entity
+        if entity:
+            self.permission_group = entity.permission_group
+        self.tags = entity.tags
+        self.save(update_fields=["permission_group", "entity", "tags"])
+        post_permission_changed.send(sender=type(self), instance=self)
+
+    def validate_change_collection(self, collection):
         """Raise validation error if data object cannot change collection."""
-        if self.entity:
+        if self.entity and self.entity.collection != collection:
             raise ValidationError(
                 "If Data is in entity, you can only move it to another collection "
                 "by moving entire entity."
             )
+        if collection is None:
+            raise ValidationError("Data object can not be removed from the container.")
 
     def _render_name(self):
         """Render data name.
@@ -656,6 +653,10 @@ class Data(BaseModel):
 
         self.name = name
 
+    def get_resource_limits(self):
+        """Return the resource limits for this data."""
+        return self.process.get_resource_limits(self)
+
 
 class DataDependency(models.Model):
     """Dependency relation between data objects."""
